# -*- coding: utf-8 -*-

import requests
from datetime import datetime, timedelta
from xml.etree import ElementTree as ET

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import transaction
from django.db.models import Q
from django.core.mail import send_mail

from edt.backends.edt_ade import _set_timetable, _open_edt_ade_shelf
from edt.models import EdtAdeGroup
from edt.timetable import Category, TimetableFactory
from portailuser.models import Ecole

# https://urllib3.readthedocs.org/en/latest/security.html#insecurerequestwarning
from requests.packages import urllib3

urllib3.disable_warnings()

HTTP_TIMEOUT = 300

EDTS = [
    (Ecole.a7, settings.EDT_AUTH.get("a7", (None, None))),
    (Ecole.ensat, settings.EDT_AUTH.get("ensat", (None, None))),
    (Ecole.n7, settings.EDT_AUTH.get("n7", (None, None))),
]

URL = "https://{ndd}.fr/jsp/webapi"
URL_ECOLE = {
    Ecole.a7.pk: "edt.inp-toulouse",
    Ecole.ensat.pk: "edt.inp-toulouse",
    Ecole.n7.pk: "edt.inp-toulouse",
}

PROJECT_ID = {
    Ecole.a7.pk: settings.EDT_PROJECT_ID.get("a7", None),
    Ecole.ensat.pk: settings.EDT_PROJECT_ID.get("ensat", None),
    Ecole.n7.pk: settings.EDT_PROJECT_ID.get("n7", None),
}
INDIV_CATEGORY = {
    Ecole.ensat.pk: "instructor",
    Ecole.n7.pk: "category8",  # |instructor',
    Ecole.a7.pk: "instructor",
}
DECALAGE_HORAIRE = {
    Ecole.a7.pk: settings.EDT_DEC_HORAIRE.get("a7", 0),
    Ecole.ensat.pk: settings.EDT_DEC_HORAIRE.get("ensat", 0),
    Ecole.n7.pk: settings.EDT_DEC_HORAIRE.get("n7", 0),
}

# NB_WEEKS = 7  # nombre de semaines


class Command(BaseCommand):
    help = "Met à jour l'edt via l'API web d'ADE"

    def add_arguments(self, parser):
        parser.add_argument(
            "--no-update-group-list",
            action="store_false",
            dest="update_list",
            default=True,
            help="Prevent updating the group list",
        )
        parser.add_argument(
            "--max-errors",
            action="store",
            type=int,
            dest="max_errors",
            default=1,
            help="Max number of errors before aborting",
        )
        parser.add_argument(
            "--school",
            action="store",
            type=str,
            dest="school",
            default="inp",
            help="School to update",
        )

    def handle(self, *args, **options):
        self.verbosity = int(options.get("verbosity", "0"))
        self.remaining_errors = options.get("max_errors", 1)
        self.update_list = options.get("update_list", True)
        self.school_i = options.get("school", "inp")

        self.main()

    def main(self):
        # 3 écoles: 276 + 64 + 77 + 3*(1+1+1) = 426 req
        # optimale: 3*(1 + 1 + 1 + 1 + 1) req
        for ecole, (username, password) in EDTS:
            if self.school_i != "inp" and ecole.ident != self.school_i:
                continue
            if username is None and password is None:
                self._debug("Pas d'identifiants disponibles pour {}", ecole)
                continue

            self._debug("École: {}", ecole)
            # connect
            url = URL.format(ndd=URL_ECOLE[ecole.pk])
            auth = {
                "function": "connect",
                "login": username,
                "password": password,
            }
            try:
                req = requests.get(url, params=auth, timeout=HTTP_TIMEOUT, verify=False)
                req.raise_for_status()
            except (
                requests.exceptions.ReadTimeout,
                requests.exceptions.HTTPError,
            ) as e:
                message = f"La mise à jour de l'edt de {ecole} a échoué. Connexion à {url} impossible.\n{e}"
                sujet = f"Échec mise à jour edt {ecole}"
                send_mail(
                    sujet,
                    message,
                    "portail@bde.inp-toulouse.fr",
                    ["ade@bde.inp-toulouse.fr"],
                )

                print(sujet, "\n", e)
                continue

            tree = ET.fromstring(req.content)
            sessionId = tree.attrib["id"]

            # set project id
            self.set_project(url, ecole, sessionId)

            # update groups
            if self.update_list:
                self._debug("update groups")
                groups = self.get_groups(url, sessionId, ecole)
            else:
                groups = EdtAdeGroup.objects.filter(ecole=ecole)
            self._debug("nb groups: {}", len(groups))

            # update edt
            decalage_horaire = timedelta(0, 3600 * DECALAGE_HORAIRE[ecole.pk])
            self._debug("update events")
            self.activities_xml = self.get_activities(url, sessionId, groups)
            for i, group in enumerate(groups):
                self._debug("group {}", i)
                self.update_events(url, sessionId, group, decalage_horaire)

            # deconnect
            deco = {
                "sessionId": sessionId,
                "function": "disconnect",
            }
            req = requests.get(url, params=deco, timeout=HTTP_TIMEOUT, verify=False)
            # req.raise_for_status()
            self._debug("FIN")

    def _debug(self, msg, *args):
        if self.verbosity:
            print(msg.format(*args))

    def set_project(self, url, ecole, sessionId):
        """Fixe l'id du projet.

        :param str url: url de l'edt
        :param Ecole ecole: école concernée
        :param str sessionId: id de la session
        """
        # init
        project = {
            "sessionId": sessionId,
        }

        # récupère l'id du projet
        project["function"] = "getProjects"
        req = requests.get(url, params=project, timeout=HTTP_TIMEOUT, verify=False)
        req.raise_for_status()

        projects_xml = ET.fromstring(req.content)
        if len(projects_xml):
            projects_id = [p.get("id") for p in projects_xml]
            if PROJECT_ID[ecole.pk] not in projects_id:
                self._debug(
                    "Error: project '{}' not in ADE projects list for {} (valid: {})",
                    PROJECT_ID[ecole.pk],
                    ecole,
                    ",".join(projects_id),
                )

        # fixe l'id du projet
        project["function"] = "setProject"
        project["projectId"] = PROJECT_ID[ecole.pk]

        req = requests.get(url, params=project, timeout=HTTP_TIMEOUT, verify=False)
        req.raise_for_status()

    def get_groups(self, url, sessionId, ecole, category=None):
        """Récupère et met à jour les groupes.

        :param str url: url de l'edt
        :param str sessionId: id de la session
        :param Ecole ecole: école concernée
        :param category: catégorie
        :return: liste des groupes
        :rtype: array[EdtAdeGroup]
        """
        # GET: function getResources
        resources = {
            "sessionId": sessionId,
            "function": "getResources",
            "detail": 11,  # n7: 5?
        }
        if category is None:
            resources["category"] = "trainee"
        else:
            resources["category"] = category
        req = requests.get(url, params=resources, timeout=HTTP_TIMEOUT, verify=False)
        req.raise_for_status()

        # init new groups
        new_groups = []
        new_groups_id = set()
        new_groups_name = dict()

        groups_avoid = ["tous", "conseil", "conf", "jury"]

        # get only non-groups resources
        groups_xml = ET.fromstring(req.content)
        groups_xml_l = groups_xml.findall(".//*[@isGroup='false']")
        for group in groups_xml_l:
            name = group.attrib["name"]
            new_groups_name[name] = name in new_groups_name
        for group in groups_xml_l:
            # name non utilisable car doublons…
            # name = '/'.join(group.attrib['path'].split('.'))
            name = group.attrib["name"]
            if ecole == Ecole.ensat:
                path = " — ".join(group.attrib["path"].split(".")[:-1])
                name = " — ".join([path, name]) if name else path
            else:
                if new_groups_name.get(name):
                    path = " — ".join(group.attrib["path"].split(".")[1:-1])
                    name = f"{name} ({path})" if name else path

            # avoid some groups
            avoid = name == ""
            for txt in groups_avoid:
                if txt in name.lower():
                    avoid = True
                    break
            if avoid:
                continue
            group_id = int(group.attrib["id"])

            self._debug("Updating group id {} ({})", group_id, name)
            new_groups.append(EdtAdeGroup(ecole=ecole, gid=group_id, name=name))
            new_groups_id.add(group_id)

        # update
        self.update_groups(ecole, new_groups)

        return sorted(new_groups, key=lambda item: item.gid)
        # return new_groups

    @transaction.atomic
    def update_groups(self, ecole, ade_groups):
        """
        Met à jour les groupes.

        :param Ecole ecole: école concernée
        :param EdtAdeGroup ade_groups: groupes récupérés depuis ADE
        """
        self._debug("Updating group list for {}", ecole)

        # On ne garde que les groupes en base dont le nom et l’ID correspondent
        # aux groupes récupérés sur ADE. Ceci va provoquer la suppression
        # (et désinscription) des groupes en base dont l’ID change suite à
        # un changement d’année.

        db_groups = EdtAdeGroup.objects.filter(ecole=ecole).exclude(gid=None)
        db_group_idents = set((group.gid, group.name) for group in db_groups)
        ade_group_idents = set((group.gid, group.name) for group in ade_groups)

        added_group_idents = ade_group_idents - db_group_idents
        deleted_group_idents = db_group_idents - ade_group_idents

        added_groups = (group for group in ade_groups if (group.gid, group.name) in added_group_idents)

        added_count = len(added_group_idents)
        removed_count = len(deleted_group_idents)
        deleted_group_ids = (group[0] for group in deleted_group_idents)

        EdtAdeGroup.objects.filter(Q(ecole=ecole) & Q(gid__in=deleted_group_ids)).delete()

        EdtAdeGroup.objects.bulk_create(added_groups)

        if added_count or removed_count:
            # Always display this so we know if the group list accessor
            # messed up
            self._debug(
                "Group list updated, {} added, {} removed.\n",
                added_count,
                removed_count,
            )
        else:
            self._debug("Group list updated, no changes.")

    def get_activities(self, url, sessionId, groups):
        """Récupère les activités liées aux événements.

        :param str url: url de base
        :param str sessionId: id de la session
        :param list(EdtAdeGroup) groups: groupes dont on veut les événements
        :return: activités au format XML
        """
        groups_id = [str(group.gid) for group in groups]

        # GET: function getActivities
        # get name, type, isActive, owner
        activities = {
            "sessionId": sessionId,
            "function": "getActivities",
            "resources": "|".join(groups_id),
            "detail": 17,
        }
        url_ = url + "?" + "&".join(k + "=" + str(v) for k, v in activities.items())
        req = requests.get(url_)
        activities_xml = ET.fromstring(req.content)

        return activities_xml

    def update_events(self, url, sessionId, group, decalage_horaire):
        """Récupère et met à jour les événements d'un groupe.

        :param str url: url de l'edt
        :param str sessionId: id de la session
        :param EdtAdeGroup group: groupe dont on veut les événements
        :param activities_xml: activités au format XML
        :param decalage_horaire: un décalage horaire à appliquer à chaque évenement
        """
        group_id = group.gid
        group_name = group.name

        # GET: function getEvents
        # get events : id, name, activityId, startHour, endHour, date, resources()
        # XXX: les semaines commencent un peu avant la rentrée, suivant les écoles…
        # week = datetime.isocalendar(datetime.today())[1]
        # weeks = '|'.join(str(k) for k in range(week - 1, week + NB_WEEKS))
        events = {
            "sessionId": sessionId,
            "function": "getEvents",
            "resources": group_id,
            # 'weeks': weeks,
            "detail": 8,
        }

        url_ = url + "?" + "&".join(k + "=" + str(v) for k, v in events.items())
        req = requests.get(url_)
        events_xml = ET.fromstring(req.content)
        nb_events = len(events_xml)
        self._debug("nb events for group {}: {}", group_name, nb_events)

        # get all activities for events
        activities_id = []
        for n_event in range(nb_events):
            activities_id.append(events_xml[n_event].attrib["activityId"])

        factory = TimetableFactory(group_name)
        for event_xml in events_xml:
            activity_id = event_xml.get("activityId")
            find_activity = f"./activity[@id='{activity_id}'][@isActive='true']"
            activity_xml = self.activities_xml.find(find_activity)

            if activity_xml is not None:
                event = ADEEvent()

                # dates
                dtstart = f"{event_xml.attrib['date']} {event_xml.attrib['startHour']}"
                dtend = f"{event_xml.attrib['date']} {event_xml.attrib['endHour']}"
                event.uid = event_xml.attrib["id"]
                event.dtstart = datetime.strptime(dtstart, "%d/%m/%Y %H:%M") + decalage_horaire
                event.dtend = datetime.strptime(dtend, "%d/%m/%Y %H:%M") + decalage_horaire

                # ressources (lieu, responsable)
                locations_xml = event_xml.findall('.//*[@category="classroom"]')
                event.locations = [loc_xml.get("name") for loc_xml in locations_xml]
                people_xml = event_xml.findall('.//*[@category="instructor"]')
                event.people = [instit_xml.get("name") for instit_xml in people_xml]

                # nom
                # XXX: à reformater?
                event.topic = event_xml.attrib["name"]
                # type et catégorie
                event.type, event.category = event.get_type(event_xml, activity_xml)

                factory.add_event(
                    event.dtstart,
                    event.dtend,
                    event.topic,
                    event.type,
                    event.category,
                    event.locations,
                    event.people,
                    event.uid,
                )

        with _open_edt_ade_shelf() as shelf:
            _shelf = shelf
        _set_timetable(_shelf, group.ecole.pk, int(group_id), factory.table)


class ADEEvent(object):
    """
    Un événement dans ADE correctement formaté.
    """

    __slots__ = (
        "uid",
        "dtstart",
        "dtend",
        "topic",
        "type",
        "category",
        "locations",
        "people",
    )

    def __init__(self):
        for item in self.__slots__:
            setattr(self, item, None)

    def __repr__(self):
        res = ["<ADEEvent: "]
        for i, item in enumerate(self.__slots__):
            if i > 0:
                res.append(", ")
            res.append(f"{item}={repr(getattr(self, item))}")
        res.append(">")
        return "".join(res)

    class BadEventError(Exception):
        pass

    def get_type(self, eventX, activityX):
        """Retourne le type et la catégorie d'un événement.

        :param ElementTree eventX: événement
        :param ElementTree activityX: activité contenant l'événement
        :return: type et catégorie de l'événement
        :rtype: list of strings
        """
        type_text = activityX.attrib["type"]
        category = "AUTRE"

        if type_text in {"Cours", "CM"}:
            type_text = "Cours"
            category = Category.COURS

        elif type_text in {"MATI"}:
            type_text = "Fourre-tout"
            category = Category.COURS

        elif type_text in {"Colloque", "spécifique"}:
            type_text = "Colloque"
            category = Category.COURS

        elif type_text in {"Rencontre"}:
            category = Category.COURS

        elif type_text in {"UE"}:
            type_text = "Entretien CV"
            category = Category.EXAMEN

        elif type_text in {"EXAM", "Examen", "EXAMEN", "TEST", "Epreuve"}:
            type_text = "Examen"
            category = Category.EXAMEN

        elif type_text in {"ORAL", "ORA", "Oral", "EXAMEN Oral"}:
            type_text = "Oral"
            category = Category.EXAMEN

        elif type_text in {"Soutenance", "Jury"}:
            category = Category.EXAMEN

        elif type_text in {"Projet", "BE", "TPBE", "TPPR"}:
            category = Category.PROJET

        elif type_text in {"TD", "CTD", "TDM", "CM TD", "Cours-TD", "CM&TP"}:
            category = Category.TD

        elif type_text in {"TP", "TPE", "TDTP"}:
            category = Category.TP

        elif type_text == "TUTO":
            type_text = "Tutorat"
            category = Category.AUTRE

        elif type_text in {"Réunion", "Reunion"}:
            category = Category.AUTRE

        elif type_text in {"Visite"}:
            category = Category.AUTRE

        elif type_text in {"SORTIE", "Sortie"}:
            type_text = "Sortie"
            category = Category.AUTRE

        elif type_text in {"Spécifique"}:
            type_text = eventX.attrib["name"]
            if type_text in {"E. P. S"}:
                category = Category.SPORT
            elif type_text in {"Projets / Travail personnel"}:
                category = Category.PROJET
            # else:
            #    category = Category.AUTRE

        else:
            if type_text:
                print(
                    type_text,
                    eventX.attrib["name"],
                    eventX.attrib["id"],
                    activityX.attrib["id"],
                    activityX.attrib["owner"],
                )

        return type_text, category
